part of tecfy_corev2_client_package;

class ReportDetailPage extends StatefulWidget {
  /// the report name
  final Map<String, dynamic> report;

  /// the filter properties
  final List<EntityFilter> filter;

  /// view report page with filter and print
  ReportDetailPage(this.report, this.filter);
  @override
  _ReportDetailPageState createState() => _ReportDetailPageState();
}

class _ReportDetailPageState extends TecfyState<ReportDetailPage> {
  List<Map<String, dynamic>>? _data;
  Map<String, List<Map<String, dynamic>>>? _dataGrouped;
  Map<String, Map<String, dynamic>>? _dataGroupedTotal;
  List<Property>? _filters;
  Map<String, dynamic>? _filterData;
  Map<String, dynamic>? _totals;
  int? _sortColumnIndex;
  bool _sortAcd = true;
  ByteData? _font;
  late List<ReportProperty> _props;
  late List<ReportProperty>? _groupProperties;
  late String? _groupBy;

  // late List<ReportProperty> _propsSortedForPdf;
  late ScrollController _scrollControllerH;
  late ScrollController _scrollControllerV;
  String _mode = "portrait";
  bool isEditing = false;
  late Map savedData;
  List<ReportProperty> savedList = [];

  @override
  void initState() {
    _scrollControllerV = ScrollController();
    _scrollControllerH = ScrollController();
    rootBundle.load("assets/fonts/STC-Regular.ttf").then((d) => _font = d);
    _props = widget.report['properties']
        .map<ReportProperty>((p) => ReportProperty.fromMap(p))
        .toList();
    _groupBy = widget.report['groupBy'];
    _groupProperties = widget.report['groupProperties']
        ?.map((r) => _props
            // ignore: unnecessary_cast
            .map((t) => t as ReportProperty?)
            .firstWhere((e) => e?.dbName == r, orElse: () => null))
        .where((r) => r != null)
        .map<ReportProperty>((t) => t as ReportProperty)
        .toList();
    _props = _props
        .where((element) =>
            _groupProperties?.any((p) => p.dbName == element.dbName) != true)
        .toList();
    savedData = ConfigService.getValueMap('Report_${widget.report['dbName']}');
    if (savedData.isNotEmpty) {
      savedData.entries.forEach((element) {
        savedList.add(ReportProperty(
            dbName: element.value['dbName'],
            nameEn: element.value['nameEn'],
            nameAr: element.value['nameAr'],
            type: ValueTypes.values[element.value['type'] ?? 0]));
      });
      savedList.remove(savedList[savedList.length - 1]);
      savedList.remove(savedList.last);
    } else {
      savedList = _props;
    }

    _getData();

    super.initState();
  }

  @override
  void dispose() {
    _scrollControllerH.dispose();
    _scrollControllerV.dispose();
    super.dispose();
  }

  void _onSort(String dbName, ValueTypes type) {
    if (_dataGrouped != null && _groupProperties != null) {
      for (var d in _dataGrouped!.values) {
        d.sort((a, b) => _sortAcd
            ? (type == ValueTypes.Number
                ? (a[dbName] - b[dbName]).toInt()
                : a[dbName].toString().compareTo(b[dbName].toString()))
            : (type == ValueTypes.Number
                ? (b[dbName] - a[dbName]).toInt()
                : b[dbName].toString().compareTo(a[dbName].toString())));
      }
    } else {
      _data?.sort((a, b) => _sortAcd
          ? (type == ValueTypes.Number
              ? (a[dbName] - b[dbName]).toInt()
              : a[dbName].toString().compareTo(b[dbName].toString()))
          : (type == ValueTypes.Number
              ? (b[dbName] - a[dbName]).toInt()
              : b[dbName].toString().compareTo(a[dbName].toString())));
    }
    setState(() {});
  }

  Future _getData() async {
    try {
      //List<dynamic> data;
      // if (ConnectionService.isOfflineMode) {
      //   var result = await EntityService().getData(
      //       EntityService.getEntity(widget.report['entityName']),
      //       filter: widget.filter,
      //       allRecords: true);
      //   data = result.data;
      // } else {
      if (TecfyCoreApp.reportService != null &&
          TecfySyncService.isOfflineMode &&
          widget.report['isOnlineOnly'] == false) {
        _data = (await TecfyCoreApp.reportService
                ?.getReportData(widget.report['dbName'], widget.filter))
            as List<Map<String, dynamic>>?;
      } else {
        var result = await TecfyBasicApp.apiService.httpGetDynamic(
            'admin/setup/Report/GetData/${widget.report['dbName']}?extraData=true' +
                (widget.filter.length > 0
                    ? '&filter=${jsonEncode(widget.filter.map<Map>((f) => f.toJson()).toList(), toEncodable: EntityService.customEncode)}'
                    : ''));
        if (!result.success) {
          await result.showMessage(context);
          return;
        }
        _data = result.data['data']
            .map<Map<String, dynamic>>((e) => e as Map<String, dynamic>)
            .toList();
        _filters = result.data['filter']
            .map<Property>((e) => Property.getProperty(e)!)
            .toList();
        _filterData = result.data?['filterData'] as Map<String, dynamic>?;
      }
      // }
      if (_data != null) {
        // if (widget.report['totals'] != null &&
        //     widget.report['totals'].length > 0) {
        //   totals = {};
        //   for (var t in widget.report['totals']) {
        //     totals[t] = 0;
        //     for (var d in data) {
        //       totals[t] += d[t] ?? 0;
        //     }
        //   }
        // }

        if (_groupBy != null) {
          _dataGrouped ??= {};
          for (var d in _data ?? []) {
            _dataGrouped![d[_groupBy]] ??= [];
            _dataGrouped![d[_groupBy]]!.add(d);
          }
          _dataGroupedTotal ??= {};
          for (var d in _dataGrouped!.values) {
            var last = d.removeLast();
            _dataGroupedTotal![last[_groupBy]] = last;
          }
        }

        _totals = _data?.removeLast();

        setState(() {});
      } else {
        await AppNavigator.showMessage(
            context, "Error or No Data!", MessageType.error);
        Navigator.pop(context);
      }
    } catch (err, trace) {
      print('XXXXXXXXXXXXXXXXXXX report load error');
      print(err);
      print(trace);
      await AppNavigator.showMessage(
          context, "Error getting data!", MessageType.error);
      Navigator.pop(context);
    }
  }

  List<DataRow> _getRows(data, [Map<String, dynamic>? totals]) {
    var result = data!
        .map<DataRow>((d) => DataRow(
              cells: savedList.map<DataCell>((p) => _getCell(p, d)).toList(),
            ))
        .toList();

    result.add(DataRow(
      cells: savedList
          .map<DataCell>((p) => _getCell(p, totals ?? _totals, totals: true))
          .toList(),
    ));
    return result;
  }

  bool checkPropInSavedTable(e) {
    late bool isContains;
    for (int i = 0; i < savedList.length; i++) {
      if (savedList[i].dbName == e.dbName) {
        isContains = true;
        break;
      } else {
        isContains = false;
      }
    }
    return isContains;
  }

  @override
  Widget build(BuildContext context) {
    return BasePage(
        context,
        (MapService.getValue(widget.report, 'name') as Map<String, dynamic>)[
            LocalConfigService.languagePrefix.toLowerCase()],
        'report',
        isMobile: LayoutService.isMobile(context),
        appBarActions: [
          AppButtonIconText(
            label: 'Print',
            icon: Icon(Icons.print, color: Colors.white),
            onPressed: _onPrint,
          ),
          SizedBox(width: 5),
          AppButtonIconText(
            label: 'PDF',
            icon: Icon(Icons.picture_as_pdf, color: Colors.white),
            onPressed: _onSharePDF,
          ),
          PopupMenuButton<String>(
            onSelected: (val) async {
              setState(() => _mode = val);
              if (val == 'Show All Properties') {
                await AppNavigator.showAppDialogEx(
                    context,
                    () => Column(
                          children: _props
                              .map((e) => Padding(
                                  padding: EdgeInsets.all(10),
                                  child: Row(
                                      mainAxisAlignment:
                                          MainAxisAlignment.spaceAround,
                                      children: [
                                        Text(e.name),
                                        Spacer(),
                                        Checkbox(
                                            value: checkPropInSavedTable(e),
                                            onChanged: (val) async {
                                              print(val);
                                              if (val == true) {
                                                print('Adding');
                                                AppNavigator.pop(context);
                                                savedList.insert(
                                                    _props.indexOf(e), e);
                                              } else {
                                                print('Removing Column');
                                                AppNavigator.pop(context);
                                                savedList.remove(e);
                                              }
                                              bool confirm = await AppNavigator
                                                  .confirmMessage(
                                                      context,
                                                      'هل تريد '
                                                      'حفظ '
                                                      'الجدول بعد التعديل ؟');
                                              if (confirm) {
                                                ReportDataService
                                                    .saveTableColumnSort(
                                                        isEditing,
                                                        savedList,
                                                        context,
                                                        widget.report,
                                                        _sortColumnIndex ?? 0,
                                                        _sortAcd);
                                              }
                                              checkPropInSavedTable(e);
                                              setState(() {});
                                            }),
                                      ])))
                              .toList(),
                        ));
              }
            },
            itemBuilder: (BuildContext context) => [
              PopupMenuItem<String>(
                value: 'portrait',
                child: Row(
                    mainAxisAlignment: MainAxisAlignment.spaceBetween,
                    children: [
                      Checkbox(
                        value: _mode == 'portrait' ? true : false,
                        onChanged: (value) {},
                        activeColor: Colors.black54,
                      ),
                      Text("portrait",
                          style: TextStyle(
                              fontSize: 15, fontWeight: FontWeight.bold)),
                    ]),
              ),
              PopupMenuItem<String>(
                value: 'landscape',
                child: Row(
                    mainAxisAlignment: MainAxisAlignment.spaceBetween,
                    children: [
                      Checkbox(
                        value: _mode == 'landscape' ? true : false,
                        onChanged: (value) {},
                        activeColor: Colors.black54,
                      ),
                      Text("Landscape",
                          style: TextStyle(
                              fontSize: 15, fontWeight: FontWeight.bold)),
                    ]),
              ),
              PopupMenuItem<String>(
                value: 'Edit Columns',
                onTap: () async {
                  isEditing = !isEditing;
                  print(isEditing);
                  if (!isEditing) {
                    savedData = await ReportDataService.saveTableColumnSort(
                        isEditing,
                        savedList,
                        context,
                        widget.report,
                        _sortColumnIndex ?? 0,
                        _sortAcd);
                    // savedList=[];
                    // if(isEditing){
                    //   savedData.entries.forEach((element) {
                    //     print(element.value['type']);
                    //     savedList.add(ReportProperty
                    //       (dbName: element.value['dbName'],nameEn: element.value['nameEn'],
                    //         nameAr:
                    //         element.value['nameAr'],type: ValueTypes.values[element
                    //             .value['type']??0]));
                    //   });
                    //   savedList.remove(savedList.last);
                    //   savedList.forEach((element) {
                    //     print('${element.dbName} -- ${element.nameEn} -- ${element.type}');
                    //   });
                    // }

                    setState(() {});
                  }
                },
                child: Row(mainAxisAlignment: MainAxisAlignment.end, children: [
                  Text(isEditing ? "Save Sort" : "Edit Columns",
                      style:
                          TextStyle(fontSize: 15, fontWeight: FontWeight.bold)),
                ]),
              ),
              PopupMenuItem<String>(
                value: 'Show All Properties',
                child: Row(mainAxisAlignment: MainAxisAlignment.end, children: [
                  Text("Show All Properties",
                      style:
                          TextStyle(fontSize: 15, fontWeight: FontWeight.bold)),
                ]),
              ),
              PopupMenuItem<String>(
                value: 'Clear all Settings',
                onTap: () async {
                  savedList = _props;
                  await ConfigService.setValueMap(
                      'Report_${widget.report['dbName']}', {});
                },
                child: Row(mainAxisAlignment: MainAxisAlignment.end, children: [
                  Text("Clear all Settings",
                      style:
                          TextStyle(fontSize: 15, fontWeight: FontWeight.bold)),
                ]),
              ),
            ],
          ),
        ],
        child: Column(
          children: [
            _getFilterTable(),
            Expanded(child: _pageContent()),
          ],
        ));
  }

  Future<dynamic> _onPrint() async {
    // var resultData = await EntityService.getData(
    //   EntityService.getEntity("Branch"),
    // );
    // if (resultData.data == null) return;
    // var branchLogo =
    //     await ImageService.getImageBytes(resultData.data![0]["logoId"]);
    var doc = _getPrintablePdf(
        // resultData.data![0]["printTitleAr"] ?? '',
        // resultData.data![0]["printSubTitleAr"] ?? '',
        // resultData.data![0]["printFooterAr"] ?? '',
        // resultData.data![0]["printSubFooterAr"] ?? '',
        // branchLogo
        "",
        "",
        "",
        "",
        null);
    var result = await Printing.layoutPdf(
      onLayout: (format) async => doc.save(),
      //name:widget.report['name${LocalConfigService.languagePrefex}'],
    );
    if (kDebugMode) print(result);
  }

  Future<dynamic> _onSharePDF() async {
    // var resultData = await EntityService.getData(
    //   EntityService.getEntity("Branch"),
    // );
    // if (resultData.data == null) return;
    // var branchLogo =
    //     await ImageService.getImageBytes(resultData.data![0]["logoId"]);
    var doc = _getPrintablePdf(
        // resultData.data![0]["printTitleAr"] ?? '',
        // resultData.data![0]["printSubTitleAr"] ?? '',
        // resultData.data![0]["printFooterAr"] ?? '',
        // resultData.data![0]["printSubFooterAr"] ?? '',
        // branchLogo,
        "",
        "",
        "",
        "",
        null);
    var result = await Printing.sharePdf(
        bytes: await doc.save(),
        filename:
            '${widget.report['name'][LocalConfigService.languagePrefix.toLowerCase()]}.pdf');
    // var result = await Printing.layoutPdf(
    //   onLayout: (format) async => doc.save(),
    //   //name:widget.report['name${LocalConfigService.languagePrefex}'],
    // );
    if (kDebugMode) print(result);
  }

  pw.Document _getPrintablePdf(
      var titleAr, var subTitleAr, var footerAr, var subFooterAr, var logoId) {
    final doc = pw.Document();
    DateTime printingDate = DateTime.now();
    String formattedDate =
        intl.DateFormat('yyyy-MM-dd   kk:mm').format(printingDate);
    doc.addPage(
      pw.MultiPage(
        orientation: _mode == "landscap"
            ? pw.PageOrientation.landscape
            : pw.PageOrientation.portrait,
        pageFormat: PdfPageFormat.a4,
        textDirection: pw.TextDirection.rtl,
        margin: pw.EdgeInsets.all(5),
        footer: (context) =>
            pw.Row(mainAxisAlignment: pw.MainAxisAlignment.center, children: [
          pw.Row(children: [
            pw.Text(" ${formattedDate.toString()} ",
                style: pw.TextStyle(font: pw.Font.ttf(_font!)),
                textDirection: pw.TextDirection.ltr),
            pw.Text('تاريخ الطباعة : ',
                style: pw.TextStyle(font: pw.Font.ttf(_font!)),
                textDirection: pw.TextDirection.rtl),
          ]),
          pw.SizedBox(width: 40),
          pw.Row(children: [
            pw.Text(" ${LocalConfigService.user!.fullName.toString()} ",
                style: pw.TextStyle(font: pw.Font.ttf(_font!)),
                textDirection: pw.TextDirection.rtl),
            pw.Text('أسم المستخدم : ',
                style: pw.TextStyle(font: pw.Font.ttf(_font!)),
                textDirection: pw.TextDirection.rtl),
          ]),
        ]),
        build: (c) => [
          pw.Padding(
              padding: pw.EdgeInsets.only(top: 0),
              child: pw.Column(children: [
                pw.Row(children: [
                  pw.Expanded(
                      child: pw.Row(children: [
                    logoId == null
                        ? pw.Container(width: 120, height: 120)
                        : pw.Container(
                            width: 120,
                            height: 120,
                            decoration: pw.BoxDecoration(
                                image: pw.DecorationImage(
                                    image: pw.MemoryImage(logoId),
                                    fit: pw.BoxFit.contain)),
                          ),
                    pw.Container(width: 40),
                  ])),
                  pw.SizedBox(width: 20),
                  pw.Expanded(
                      child: pw.Column(
                    crossAxisAlignment: pw.CrossAxisAlignment.end,
                    children: [
                      pw.SizedBox(height: 5),
                      pw.Text(
                        titleAr.toString(),
                        style: pw.TextStyle(
                            font: pw.Font.ttf(_font!),
                            fontWeight: pw.FontWeight.bold,
                            fontSize: 16),
                        // textDirection: pw.TextDirection.rtl
                      ),
                      pw.Text(subTitleAr.toString(),
                          style: pw.TextStyle(font: pw.Font.ttf(_font!)),
                          textDirection: pw.TextDirection.rtl),
                      pw.Text(footerAr.toString(),
                          style: pw.TextStyle(font: pw.Font.ttf(_font!)),
                          textDirection: pw.TextDirection.rtl),
                      pw.Text(subFooterAr.toString(),
                          style: pw.TextStyle(font: pw.Font.ttf(_font!)),
                          textDirection: pw.TextDirection.rtl),
                    ],
                  )),
                ]),
              ])),
          pw.Center(
              child: pw.Text(
                  widget.report['name${LocalConfigService.languagePrefix}'],
                  style:
                      pw.TextStyle(font: pw.Font.ttf(_font!), fontSize: 15))),
          pw.SizedBox(height: 20),
          _getFilterTablePdf(),
          _dataGrouped != null && _groupProperties != null
              ? pw.Column(
                  children: _dataGrouped?.values
                          .map((e) => pw.Column(
                                  mainAxisSize: pw.MainAxisSize.min,
                                  children: [
                                    pw.SizedBox(height: 20),
                                    pw.Row(
                                      mainAxisAlignment:
                                          pw.MainAxisAlignment.start,
                                      children: _groupProperties!
                                          .map(
                                            (p) => pw.Padding(
                                              padding: const pw.EdgeInsets.only(
                                                  left: 20.0, right: 20.0),
                                              child: pw.Row(
                                                crossAxisAlignment:
                                                    pw.CrossAxisAlignment.start,
                                                mainAxisAlignment:
                                                    pw.MainAxisAlignment.start,
                                                children: [
                                                  _getCellReport(p, e.first),
                                                  pw.SizedBox(width: 5),
                                                  pw.Text(':'),
                                                  pw.SizedBox(width: 5),
                                                  pw.Text(p.nameAr,
                                                      style: pw.TextStyle(
                                                          font: pw.Font.ttf(
                                                              _font!)),
                                                      maxLines: 1),
                                                ],
                                              ),
                                            ),
                                          )
                                          .toList(),
                                    ),
                                    _getDataTablePdf(
                                        e,
                                        e.length == 0
                                            ? null
                                            : _dataGroupedTotal![
                                                e.first[_groupBy]]),
                                  ]))
                          .toList() ??
                      [])
              : _getDataTablePdf(_data),
        ],
      ),
    );

    return doc;
  }

  pw.Widget _getDataTablePdf(List<Map<String, dynamic>>? data,
      [Map<String, dynamic>? total]) {
    return pw.Table(
        //columnWidths: sortColumnIndex,
        border: pw.TableBorder(
            horizontalInside: pw.BorderSide(),
            top: pw.BorderSide(),
            bottom: pw.BorderSide(),
            right: pw.BorderSide(),
            left: pw.BorderSide()),
        children: (data
                ?.map<pw.TableRow>((d) => pw.TableRow(
                      children: savedList.reversed
                          .map<pw.Widget>((p) => _getCellReport(p, d))
                          .toList(),
                    ))
                .toList() ??
            [])
          ..insert(
            0,
            pw.TableRow(
              decoration: pw.BoxDecoration(color: PdfColors.grey300),
              children: savedList.reversed
                  .map<pw.Widget>(
                    (p) => pw.Padding(
                      padding: pw.EdgeInsets.only(left: 5, right: 5),
                      child: pw.Align(
                        alignment: LocalConfigService.isArabic
                            ? pw.Alignment.centerRight
                            : pw.Alignment.centerLeft,
                        child: pw.Text(
                          p.name,
                          style: pw.TextStyle(font: pw.Font.ttf(_font!)),
                        ),
                      ),
                    ),
                  )
                  .toList(),
            ),
          )
          ..add(pw.TableRow(
              decoration: pw.BoxDecoration(color: PdfColors.grey200),
              children: savedList.reversed
                  .map<pw.Widget>(
                      (p) => _getCellReport(p, total ?? _totals, totals: true))
                  .toList())));
  }

  pw.Widget _getCellReport(ReportProperty prop, data, {totals = false}) {
    var val = data[prop.dbName];
    PdfColor color = PdfColors.black;
    if (prop.colorField?.isNotEmpty ?? false) {
      try {
        color = hexToColorPdf(data[prop.colorField]);
      } catch (err) {
        print(err);
      }
    }
    if (totals) {
      if (val == null || val == 0 || val == '' || val == '0')
        return pw.Container();
    } else {
      if (val == null) return pw.Container();
    }
    var textStyle = totals
        ? pw.TextStyle(
            font: pw.Font.ttf(_font!),
            fontSize: 13,
            fontWeight: pw.FontWeight.bold)
        : pw.TextStyle(font: pw.Font.ttf(_font!), color: color);
    pw.Widget cell;
    switch (prop.type) {
      case ValueTypes.Text:
        cell = pw.Text(val, style: textStyle);
        break;
      case ValueTypes.Number:
        if (prop.format == '0')
          return pw.Text((val as num).toInt().toString(), style: textStyle);
        cell = pw.Text((val as num).toDouble().toStringAsFixed(2),
            style: textStyle);
        break;
      case ValueTypes.DateTime:
        DateTime? date;
        if (val is num)
          date = DateTime.fromMillisecondsSinceEpoch(val.toInt()).toLocal();
        if (val is String) date = BaseModel.parseUTC(val)!;
        if (val is DateTime) date = val;

        cell = pw.Text(
            intl.DateFormat(prop.format ?? "yyyy-MM-dd HH:mm")
                .format(date!.toLocal()),
            style: textStyle);

        // cell = pw.Text(val.toString().replaceFirst('T', ' ').substring(0, 16),
        //     style: textStyle);
        break;
      default:
        cell = pw.Text('-');
        break;
    }
    return pw.Padding(
        padding: pw.EdgeInsets.only(left: 5, right: 5),
        child: pw.Align(
            alignment: LocalConfigService.isArabic
                ? pw.Alignment.centerRight
                : pw.Alignment.centerLeft,
            child: cell));
  }

  pw.Widget _getFilterTablePdf() {
    if (_filters == null) return pw.Container();
    return pw.Container(
        constraints: pw.BoxConstraints(maxWidth: 350),
        child: pw.Column(
          children: _filters!
              .map((e) => pw.Row(
                    children: [
                      pw.Expanded(
                          flex: 4,
                          child: pw.Text(e.getDisplayText(_filterData ?? {}),
                              style: pw.TextStyle(font: pw.Font.ttf(_font!)))),
                      pw.Expanded(
                          flex: 3,
                          child: pw.Text(
                            e.nameAr,
                            style: pw.TextStyle(font: pw.Font.ttf(_font!)),
                          )),
                    ],
                  ))
              .toList(),
        ));
  }

  List<DataColumn> getColumns() => savedList
      .map<DataColumn>((p) => DataColumn(
            numeric: p.type == ValueTypes.Number,
            onSort: (s, d) {
              _sortColumnIndex = s;
              if (savedData['sortColumnIndex'] == null)
                savedData['sortColumnIndex'] = {'sortColumnIndex': 0};

              savedData['sortColumnIndex']['sortColumnIndex'] = s;
              _sortAcd = d;
              if (savedData['sortAcd'] == null)
                savedData['sortAcd'] = {'sortAcd': true};
              savedData['sortAcd']['sortAcd'] = d;
              _onSort(p.dbName, p.type);
            },
            tooltip: p.nameEn,
            label: isEditing
                ? Container(
                    padding: EdgeInsets.all(0),
                    margin: EdgeInsets.all(0),
                    height: 50,
                    child: Column(
                      children: [
                        Expanded(
                          child: Row(
                            children: [
                              IconButton(
                                  iconSize: 5,
                                  tooltip: LocalConfigService.isArabic
                                      ? "التحرك لليمين"
                                      : "Move Right",
                                  padding: EdgeInsets.all(0),
                                  onPressed: () {
                                    final index = savedList.indexOf(p);
                                    if (index > 0) {
                                      var temp;
                                      temp = savedList[index];
                                      savedList[index] = savedList[index - 1];
                                      savedList[index - 1] = temp;
                                      setState(() {});
                                    }
                                  },
                                  icon: Icon(
                                    Icons.play_arrow,
                                    size: 15,
                                  )),
                              IconButton(
                                  icon: Icon(
                                    Icons.hide_source,
                                    size: 15,
                                    color: Colors.red,
                                  ),
                                  onPressed: () {
                                    setState(() {
                                      savedList.remove(p);
                                    });
                                  }),
                              IconButton(
                                  iconSize: 5,
                                  tooltip: LocalConfigService.isArabic
                                      ? "التحرك لليسار"
                                      : "Move Left",
                                  padding: EdgeInsets.all(0),
                                  onPressed: () {
                                    final index = savedList.indexOf(p);
                                    if (index < savedList.length - 1) {
                                      var temp;
                                      temp = savedList[index];
                                      savedList[index] = savedList[index + 1];
                                      savedList[index + 1] = temp;
                                      setState(() {});
                                    }
                                  },
                                  icon: RotatedBox(
                                      quarterTurns: 2,
                                      child: Icon(
                                        Icons.play_arrow,
                                        size: 15,
                                      ))),
                            ],
                          ),
                        ),
                        Text(
                          p.name,
                          style: ThemeService.titleStyle.copyWith(height: 0),
                        ),
                      ],
                    ),
                  )
                : Text(
                    p.name,
                    style: ThemeService.titleStyle.copyWith(height: 0),
                  ),
          ))
      .toList();

  Widget _pageContent() {
    return _data == null
        ? Center(child: AppNavigator.getLoadingWidget(70))
        // : data != null
        //     ? PdfPreview(build: (c) => getPrintablePdf().save())
        : Scrollbar(
            // isAlwaysShown: true,
            // showTrackOnHover: true,
            controller: _scrollControllerH,
            child: SingleChildScrollView(
              controller: _scrollControllerH,
              scrollDirection: Axis.horizontal,
              child: Scrollbar(
                // isAlwaysShown: true,
                // showTrackOnHover: true,
                controller: _scrollControllerV,
                child: SingleChildScrollView(
                  controller: _scrollControllerV,
                  scrollDirection: Axis.vertical,
                  child: _dataGrouped != null && _groupProperties != null
                      ? Column(
                          children: _dataGrouped?.values
                                  .map((e) => Column(
                                          mainAxisSize: MainAxisSize.min,
                                          children: [
                                            Row(
                                              mainAxisAlignment:
                                                  MainAxisAlignment.start,
                                              children: _groupProperties!
                                                  .map(
                                                    (p) => Padding(
                                                      padding:
                                                          const EdgeInsets.only(
                                                              left: 20.0,
                                                              right: 20.0),
                                                      child: Row(
                                                        crossAxisAlignment:
                                                            CrossAxisAlignment
                                                                .start,
                                                        mainAxisAlignment:
                                                            MainAxisAlignment
                                                                .start,
                                                        children: [
                                                          Text(p.nameAr,
                                                              maxLines: 1),
                                                          SizedBox(width: 5),
                                                          Text(':'),
                                                          SizedBox(width: 5),
                                                          _getCellValue(
                                                              p, e.first),
                                                        ],
                                                      ),
                                                    ),
                                                  )
                                                  .toList(),
                                            ),
                                            _getDataTable(
                                                e,
                                                e.length == 0
                                                    ? null
                                                    : _dataGroupedTotal![
                                                        e.first[_groupBy]])
                                          ]))
                                  .toList() ??
                              [])
                      : _getDataTable(_data),
                ),
              ),
            ));
  }

  Widget _getDataTable(List<Map<String, dynamic>>? data,
      [Map<String, dynamic>? totals]) {
    return DataTable(
      showCheckboxColumn: true,
      sortColumnIndex: savedData.isNotEmpty &&
              savedData['sortColumnIndex']?['sortColumnIndex'] != null
          ? savedData['sortColumnIndex']['sortColumnIndex']
          : _sortColumnIndex,
      sortAscending:
          savedData.isNotEmpty && savedData['sortAcd']['sortAcd'] != null
              ? savedData['sortAcd']['sortAcd']
              : _sortAcd,
      columns: getColumns(),
      rows: _getRows(data, totals),
      // ..insert(0, TableRow(
      //   children: widget.report['properties'].map<TableCell>((p)=>TableCell(child: Text(p['name${LocalConfigService.languagePrefex}'] ?? 'N/A', style: ThemeService.titleStyle))).toList(),
      // )
      //)
    );
  }

  Color hexToColor(String code) {
    return new Color(int.parse(code.substring(1, 7), radix: 16) + 0xFF000000);
  }

  PdfColor hexToColorPdf(String code) {
    return new PdfColor.fromHex(code);
  }

  DataCell _getCell(ReportProperty prop, data, {totals = false}) {
    var val = data[prop.dbName];
    dynamic onTap;

    if (prop.drillDownEntity != null && prop.drillDownIdField != null) {
      onTap = () async {
        AppNavigator.showLoading(context);
        var entity = prop.drillDownEntity!;
        if (prop.drillDownEntityDynamic) entity = data[prop.drillDownEntity!];
        var d =
            await EntityService.getOne(entity, data[prop.drillDownIdField!]);
        if (d == null) {
          AppNavigator.pop(context);
          return;
        }
        await AppNavigator.showAppDialogEx(
            context,
            () => EntityFormPage(
                  entity,
                  data: d,
                  isPopup: true,
                ),
            closeButton: false,
            size: DialogBoxSizes.large);
        AppNavigator.pop(context);
        return;
      };
    }
    Color color = Colors.black;
    if (prop.colorField?.isNotEmpty ?? false) {
      try {
        color = hexToColor(data[prop.colorField]);
      } catch (err) {}
    }
    if (totals) {
      if (val == null || val == 0 || val == '' || val == '0')
        return DataCell(Container());
    } else {
      if (val == null) return DataCell(Text('.'));
    }
    switch (prop.type) {
      case ValueTypes.Text:
        return DataCell(
            Text(val,
                style: totals
                    ? ThemeService.titleStyle
                    : TextStyle(fontSize: 12, color: color)),
            onTap: onTap);
      case ValueTypes.Number:
        if (val is String) {
          val = num.parse(val);
        }
        if (prop.format == '0')
          return DataCell(Text((val as num).toInt().toString(),
              style: totals
                  ? ThemeService.titleStyle
                  : TextStyle(fontSize: 12, color: color)));
        else
          return DataCell(Text((val as num).toDouble().toStringAsFixed(2),
              style: totals
                  ? ThemeService.titleStyle
                  : TextStyle(fontSize: 12, color: color)));
      case ValueTypes.DateTime:
        DateTime? date = BaseModel.getDateEx(val);
        // if (val is num)
        //   date = DateTime.fromMillisecondsSinceEpoch(val.toInt()).toLocal();
        // if (val is String) date = BaseModel.parseUTC(val)!;
        // if (val is DateTime) date = val;

        return DataCell(Text(
            date == null
                ? ''
                : intl.DateFormat(prop.format ?? "yyyy-MM-dd HH:mm")
                    .format(date.toLocal()),
            style: totals
                ? ThemeService.titleStyle
                : TextStyle(fontSize: 10, color: color)));
      default:
        return DataCell(Text('-'));
    }
  }

  Text _getCellValue(ReportProperty prop, data, {totals = false}) {
    var val = data[prop.dbName];
    dynamic onTap;

    if (prop.drillDownEntity != null && prop.drillDownIdField != null) {
      onTap = () async {
        AppNavigator.showLoading(context);
        var entity = prop.drillDownEntity!;
        if (prop.drillDownEntityDynamic) entity = data[prop.drillDownEntity!];
        var d =
            await EntityService.getOne(entity, data[prop.drillDownIdField!]);
        if (d == null) {
          AppNavigator.pop(context);
          return;
        }
        await AppNavigator.showAppDialogEx(
            context,
            () => EntityFormPage(
                  entity,
                  data: d,
                  isPopup: true,
                ),
            closeButton: false,
            size: DialogBoxSizes.large);
        AppNavigator.pop(context);
        return;
      };
    }
    Color color = Colors.black;
    if (prop.colorField?.isNotEmpty ?? false) {
      try {
        color = hexToColor(data[prop.colorField]);
      } catch (err) {}
    }
    if (totals) {
      if (val == null || val == 0 || val == '' || val == '0') return Text('');
    } else {
      if (val == null) return Text('.');
    }
    switch (prop.type) {
      case ValueTypes.Text:
        return Text(val,
            style: totals
                ? ThemeService.titleStyle
                : TextStyle(fontSize: 12, color: color));
      case ValueTypes.Number:
        if (val is String) {
          val = num.parse(val);
        }
        if (prop.format == '0')
          return Text((val as num).toInt().toString(),
              style: totals
                  ? ThemeService.titleStyle
                  : TextStyle(fontSize: 12, color: color));
        else
          return Text((val as num).toDouble().toStringAsFixed(2),
              style: totals
                  ? ThemeService.titleStyle
                  : TextStyle(fontSize: 12, color: color));
      case ValueTypes.DateTime:
        DateTime? date = BaseModel.getDateEx(val);
        // if (val is num)
        //   date = DateTime.fromMillisecondsSinceEpoch(val.toInt()).toLocal();
        // if (val is String) date = BaseModel.parseUTC(val)!;
        // if (val is DateTime) date = val;

        return Text(
            date == null
                ? ''
                : intl.DateFormat(prop.format ?? "yyyy-MM-dd HH:mm")
                    .format(date.toLocal()),
            style: totals
                ? ThemeService.titleStyle
                : TextStyle(fontSize: 10, color: color));
      default:
        return Text('-');
    }
  }

  Widget _getFilterTable() {
    if (_filters == null) return Container();
    return Container(
        constraints: BoxConstraints(maxWidth: 350),
        child: Column(
          children: _filters!
              .map((e) => Row(
                    children: [
                      Expanded(flex: 3, child: Text(e.nameAr)),
                      Expanded(
                          flex: 4,
                          child: Text(e.getDisplayText(_filterData ?? {}))),
                    ],
                  ))
              .toList(),
        ));
  }

  // _lowerFirst(String s) {
  //   return '${s[0].toLowerCase()}${s.substring(1)}';
  // }
}
